from __future__ import absolute_import, unicode_literals

from robot.running.model import TestSuite

from .base import LibraryComponent


class LibraryListener(LibraryComponent):
    ROBOT_LISTENER_API_VERSION = 3

    def start_suite(self, data, result):
        # type: (TestSuite, TestSuite) -> None
        self._create_eyes_runner_if_needed()
        self.debug("Runner created")

    def close(self):
        # if `Eyes Get All Test Results` was called `eyes_runner` should be set to None
        if self.ctx.eyes_runner:
            self.warn(
                "Run `Eyes Get All Test Results` keyword in `Suite Teardown` "
                "explicitly to see all output logs"
            )
            test_results = self.ctx.eyes_runner.get_all_test_results()
            self.log_to_console(test_results)



from __future__ import absolute_import, unicode_literals

from typing import TYPE_CHECKING, List, Union

if TYPE_CHECKING:
    # TODO improve typing here
    from applitools.common.utils.custom_types import CssSelector

    Locator = Union[CssSelector, List[CssSelector]]  # typedef



from __future__ import absolute_import, unicode_literals

from typing import TYPE_CHECKING, Text

from appium.webdriver.common.mobileby import MobileBy
from AppiumLibrary import AppiumLibrary
from AppiumLibrary.locators import ElementFinder as AppiumElementFinder
from selenium.webdriver.remote.webelement import By
from SeleniumLibrary import SeleniumLibrary
from SeleniumLibrary.locators import ElementFinder as SeleniumElementFinder

from applitools.common.validators import is_webelement

from .base import LibraryComponent
from .errors import EyesLibraryValueError

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import BySelector


class AppiumElementFinderAdapter(AppiumElementFinder):
    def __init__(self, driver):
        super(AppiumElementFinderAdapter, self).__init__()
        self.__driver = driver


SELENIUM_LOCATOR_TO_BY_SELECTOR = {
    "id": By.ID,
    "xpath": By.XPATH,
    "link": By.LINK_TEXT,
    "partial link": By.PARTIAL_LINK_TEXT,
    "name": By.NAME,
    "tag": By.TAG_NAME,
    "class": By.CLASS_NAME,
    "css": By.CSS_SELECTOR,
}

APPIUM_LOCATOR_TO_BY_SELECTOR = {
    "accessibility_id": MobileBy.ACCESSIBILITY_ID,
    "android": MobileBy.ANDROID_UIAUTOMATOR,
    "ios": MobileBy.IOS_UIAUTOMATION,
    "nsp": MobileBy.IOS_PREDICATE,
    "chain": MobileBy.IOS_CLASS_CHAIN,
}
APPIUM_LOCATOR_TO_BY_SELECTOR.update(SELENIUM_LOCATOR_TO_BY_SELECTOR)


class LocatorConverter(LibraryComponent):
    def __init__(self, *args, **kwargs):
        super(LocatorConverter, self).__init__(*args, **kwargs)
        if isinstance(self.ctx.current_library, SeleniumLibrary):
            self._element_finder = SeleniumElementFinder(self.library)
            self._locator_to_by_selector = SELENIUM_LOCATOR_TO_BY_SELECTOR
        elif isinstance(self.ctx.current_library, AppiumLibrary):
            self._element_finder = AppiumElementFinderAdapter(self.driver)
            self._locator_to_by_selector = APPIUM_LOCATOR_TO_BY_SELECTOR
        else:
            raise EyesLibraryValueError(
                "Not supported library. Should be `SeleniumLibrary` or `AppiumLibrary`"
            )

    def to_by_selector(self, locator):
        # type: (Text) -> BySelector
        if is_webelement(locator):
            raise EyesLibraryValueError("Cannot convert WebElement to selector")
        prefix, criteria = self._element_finder._parse_locator(locator)
        if prefix is None:
            # appium returns None instead of `default`
            prefix = "default"
        if not criteria:
            raise ValueError("Incorrect selector: `{}`".format(locator))
        return self.__locator_to_by_selector(prefix, criteria)

    def __locator_to_by_selector(self, prefix, criteria):
        # type: (Text, Text) -> BySelector
        if prefix == "default":
            if criteria.startswith("//"):
                return [By.XPATH, criteria]
            return [By.ID, criteria]
        return [self._locator_to_by_selector[prefix], criteria]



from __future__ import absolute_import, unicode_literals

from typing import TYPE_CHECKING, Text

from AppiumLibrary import AppiumLibrary
from robot.api import logger as robot_logger
from robot.api.deco import keyword  # noqa
from SeleniumLibrary import SeleniumLibrary

from applitools.common.utils import cached_property
from applitools.selenium import ClassicRunner, VisualGridRunner

from .config_parser import SelectedRunner
from .errors import EyesLibraryValueError

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import AnyWebDriver, BySelector
    from applitools.selenium import Eyes

    from . import EyesLibrary
    from .custom_types import Locator

__all__ = ("ContextAware", "LibraryComponent", "keyword")


class ContextAware(object):
    def __init__(self, ctx):
        # type: (EyesLibrary) -> None
        """Base class exposing attributes from the common context.

        :param ctx: The library itself as a context object.
        :type ctx: SeleniumLibrary.SeleniumLibrary
        """
        self.ctx = ctx
        self.log = robot_logger

    def from_locator_to_supported_form(self, locator):
        # type: (Locator) -> BySelector
        return self.ctx._locator_converter.to_by_selector(locator)

    def from_locators_to_supported_form(self, locator):
        # type: (Locator) -> list[BySelector]
        """
        Returns [By.NAME, 'selector'] with Selenium and Appium locators or WebElement
        """
        if isinstance(locator, list):
            return [self.from_locator_to_supported_form(loc) for loc in locator]
        else:
            return [self.from_locator_to_supported_form(locator)]

    @property
    def driver(self):
        return self.ctx.driver

    @property
    def library(self):
        return self.ctx.current_library

    @property
    def drivers(self):
        return self.ctx._drivers

    @cached_property
    def defined_keywords(self):
        # type: () -> list[str]
        return list(self.ctx.keywords.keys())


class RobotWebRunner(ClassicRunner):
    BASE_AGENT_ID = "eyes.python.robotframework.selenium"


class RobotMobileNativeRunner(ClassicRunner):
    BASE_AGENT_ID = "eyes.python.robotframework.appium"


class RobotWebUFGRunner(VisualGridRunner):
    BASE_AGENT_ID = "eyes.python.robotframework.visual_grid"


class LibraryComponent(ContextAware):
    _selected_runner = None
    _log_level = None
    _selected_runner_to_eyes_runner = {
        SelectedRunner.web: RobotWebRunner,
        SelectedRunner.mobile_native: RobotMobileNativeRunner,
        SelectedRunner.web_ufg: RobotWebUFGRunner,
    }

    def to_by_selector(self, locator):
        return self.ctx._locator_converter.to_by_selector(locator)

    def info(self, msg, html=False):
        # type: (Text, bool) -> None
        self.log.info(msg, html)

    def debug(self, msg, html=False):
        self.log.debug(msg, html)

    def warn(self, msg, html=False):
        # type: (Text, bool) -> None
        self.log.warn(msg, html)

    def log_to_console(self, msg):
        self.log.console(msg)

    def log_source(self, loglevel="INFO"):
        # type: (Text) -> None
        self.ctx.log_source(loglevel)

    def register_eyes(self, eyes, alias=None):
        self.ctx.register_eyes(eyes, alias)

    def get_configuration(self):
        return self.ctx.configure.clone()

    @property
    def eyes_runner(self):
        # type: () -> VisualGridRunner | ClassicRunner
        if self.ctx.eyes_runner is None:
            raise RuntimeError(
                "Eyes runner is None. Need to parse configuration "
                "and initialize runner first"
            )
        return self.ctx.eyes_runner

    @property
    def selected_runner(self):
        # type: () -> SelectedRunner
        return self.ctx.selected_runner

    @property
    def current_eyes(self):
        # type: () -> Eyes
        return self.ctx.current_eyes

    def _create_eyes_runner_if_needed(self):
        # type: () -> None
        if self.ctx.eyes_runner is None:
            # TODO: probably need to add runner_options to Configuration class
            runner_options = self.ctx.configure.runner_options
            selected_runner = self._selected_runner_to_eyes_runner[
                self.ctx.selected_runner
            ]

            self.ctx.eyes_runner = (
                selected_runner(runner_options) if runner_options else selected_runner()
            )

    def fetch_driver(self):
        # type: () -> AnyWebDriver
        if isinstance(self.ctx.current_library, SeleniumLibrary):
            return self.ctx.current_library.driver
        elif isinstance(self.ctx.current_library, AppiumLibrary):
            return self.ctx.current_library._current_application()
        else:
            raise EyesLibraryValueError(
                "Not supported library. Should be `SeleniumLibrary` or `AppiumLibrary`"
            )



from __future__ import absolute_import, unicode_literals

from typing import TYPE_CHECKING, Any, Optional, Text

from appium.webdriver import WebElement as AppiumWebElement
from robot.api.deco import keyword as original_keyword
from robot.libraries.BuiltIn import BuiltIn
from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement
from six import string_types as basestring

from applitools.common import MatchResult
from applitools.common.utils import argument_guard
from applitools.selenium import Target

from ..base import LibraryComponent
from ..utils import collect_check_settings, is_webelement_guard, parse_region
from .keyword_tags import CHECK_FLOW, CHECK_SETTINGS_SUPPORT, TARGET_SUPPORT

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import AnyWebElement

    from ..custom_types import Locator


def keyword(name=None, tags=(), types=()):
    """Keyword with predefined CHECK_SETTINGS_SUPPORT tag"""
    tags = tags + (CHECK_SETTINGS_SUPPORT,)
    return original_keyword(name, tags, types)


def try_resolve_tag_and_keyword(tag, check_settings_keywords, defined_keywords):
    if tag in defined_keywords:
        check_settings_keywords = (tag,) + check_settings_keywords
        tag = None
    return check_settings_keywords, tag


class CheckKeywords(LibraryComponent):
    @keyword("Eyes Check Window", types=(str,), tags=(CHECK_FLOW,))
    def check_window(self, tag=None, *check_settings_keywords):
        # type: (Optional[Text], tuple[Any]) -> MatchResult
        """
        Check current browser window

        *Example:*
            |  Eyes Check Window   |
        """
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.window(), self.defined_keywords, *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword("Eyes Check Region By Coordinates", tags=(CHECK_FLOW,))
    def check_region_by_coordinates(
        self,
        region,  # type: Locator
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
        Check specified region

          |  =Arguments=  | =Description=                                                       |
          |  Region       | *Mandatory* - The region to check in format [left top width height] ,e.g. [100 200 300 300]  |

        *Example:*
            |  Eyes Check Region By Coordinates   |  [40 50 200 448]  |
        """
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.region(parse_region(region)),
            self.defined_keywords,
            *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword(
        "Eyes Check Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement), "tag": str},
        tags=(CHECK_FLOW,),
    )
    def check_region_by_element(
        self,
        element,  # type: Locator
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
        Check specified region by element

            |  =Arguments=  | =Description=                                   |
            | Element       | *Mandatory* - The element to check              |

        *Example:*
            |  Eyes Check Region By Element  |  ${element}  |
        """
        is_webelement_guard(element)
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.region(element), self.defined_keywords, *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword(
        "Eyes Check Region By Selector",
        types=(str, str),
        tags=(CHECK_FLOW,),
    )
    def check_region_by_selector(
        self,
        selector,  # type: Locator
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
        Check specified region by selector

            | =Arguments=   | =Description=                         |
            |  Selector     | *Mandatory* - The selector to check.  |

        *Example:*
            |  Eyes Check Region By Element  |  css:#selector  |
        """
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.region(self.from_locator_to_supported_form(selector)),
            self.defined_keywords,
            *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword(
        "Eyes Check Frame By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement), "tag": str},
        tags=(CHECK_FLOW,),
    )
    def check_frame_by_element(
        self,
        element,  # type: AnyWebElement
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
         Check specified frame by element

            | =Arguments=   | =Description=                                        |
            | Element       | *Mandatory* - The frame to check                     |

        *Example:*
            |  Eyes Check Frame By Element  |  ${element}  |
        """
        is_webelement_guard(element)
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.frame(element), self.defined_keywords, *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword("Eyes Check Frame By Index", types=(int, str), tags=(CHECK_FLOW,))
    def check_frame_by_index(
        self,
        frame_index,  # type: int
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
         Check specified frame by index

            | =Arguments=     | =Description=                                                       |
            |  Frame Index    | *Mandatory* - Index of the frame to check. |

        *Example:*
            |  Eyes Check Frame By Index  |  2  |
        """
        argument_guard.is_a(frame_index, int)
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.frame(frame_index), self.defined_keywords, *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword("Eyes Check Frame By Name", types=(str, str), tags=(CHECK_FLOW,))
    def check_frame_by_name(
        self,
        frame_name,  # type: Text
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
         Check specified frame by name

            |  =Arguments=   | =Description=                                   |
            |  Frame Name    | *Mandatory* - Name of the frame to check.      |

        *Example:*
            |  Eyes Check Frame By Name  |  frameName  |
        """
        argument_guard.is_a(frame_name, basestring)
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.frame(frame_name), self.defined_keywords, *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword("Eyes Check Frame By Selector", types=(str, str), tags=(CHECK_FLOW,))
    def check_frame_by_selector(
        self,
        selector,  # type: Text
        tag=None,  # type: Optional[Text]
        *check_settings_keywords  # type: tuple[Any]
    ):
        # type: (...) -> MatchResult
        """
         Check specified frame by name

            |  =Arguments=   | =Description=                                  |
            |  Selector     | *Mandatory* - Selector of the frame to check.   |

        *Example:*
            |  Eyes Check Frame By Selector  |  css:#selector   |
        """
        argument_guard.is_a(selector, basestring)
        check_settings_keywords, tag = try_resolve_tag_and_keyword(
            tag, check_settings_keywords, self.defined_keywords
        )
        check_settings = collect_check_settings(
            Target.frame(self.from_locator_to_supported_form(selector)),
            self.defined_keywords,
            *check_settings_keywords
        )
        return self.current_eyes.check(check_settings, tag)

    @keyword("Eyes Check", tags=(TARGET_SUPPORT, CHECK_FLOW))
    def check(self, target_keyword, *check_settings_keywords):
        """
         Check with target

            |  =Arguments=      | =Description=                  |
            |  Target Keyword  | *Mandatory* - Target Keyword that market with Target Keyword tag  |

        *Example:*
            |  Eyes Check  |  Target Window   |
            |  Eyes Check  |  Target Region By Coordinates   |  [34 56 78 89]  |
        """
        target = BuiltIn().run_keyword(target_keyword, *check_settings_keywords)
        self.current_eyes.check(target)



from __future__ import absolute_import, unicode_literals

import itertools
import os
from enum import Enum
from typing import Text, Type

import trafaret as trf
import yaml
from six import raise_from

from applitools.common import (
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    DeviceName,
    IosDeviceInfo,
    IosDeviceName,
    IosVersion,
    ProxySettings,
    RectangleSize,
    ScreenOrientation,
    StitchMode,
    VisualGridOption,
)
from applitools.common.selenium import BrowserType
from applitools.selenium import BatchInfo, RunnerOptions

from .config import RobotConfiguration
from .errors import EyesLibraryConfigError, EyesLibraryValueError
from .utils import get_enum_by_name, parse_viewport_size


class SelectedRunner(Enum):
    web = "web"
    web_ufg = "web_ufg"
    mobile_native = "mobile_native"


class RobotStitchMode(Enum):
    CSS = StitchMode.CSS
    SCROLL = StitchMode.Scroll


class ToEnumTrafaret(trf.Trafaret):
    def __init__(self, convert_to_enum):
        # type: (Type[Enum])->None
        self.converter = convert_to_enum

    def check_and_return(self, value, context=None):
        try:
            return get_enum_by_name(value, self.converter)
        except ValueError:
            raise trf.DataError(
                "Incorrect value `{val}`. Possible variants: {possible_vals}".format(
                    val=value,
                    possible_vals=", ".join(e.name for e in self.converter),
                ),
                value=value,
                trafaret=self,
            )


class BatchInfoTrafaret(trf.Trafaret):
    scheme = trf.Dict(
        {
            trf.Key("id", optional=True): trf.String,
            trf.Key("name", optional=True): trf.String,
            trf.Key("batch_sequence_name", optional=True)
            >> "sequence_name": trf.String,
            trf.Key("started_at", optional=True): trf.DateTime,
            trf.Key("properties", optional=True): trf.List(
                trf.Dict(name=trf.String, value=trf.String)
            ),
        },
    )

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context=context)
        batch = BatchInfo()
        for key, val in sanitized.items():
            setattr(batch, key, val)
        return batch


class ViewPortTrafaret(trf.Trafaret):
    scheme = trf.Dict(width=trf.Int, height=trf.Int) | parse_viewport_size

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return RectangleSize.from_(sanitized)


class VisualGridOptionsTrafaret(trf.Trafaret):
    scheme = trf.List(trf.Dict({"key": trf.String, "value": trf.String}))

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return [VisualGridOption(**dct) for dct in sanitized]


class DesktopBrowserInfoTrafaret(trf.Trafaret):
    scheme = trf.List(
        trf.Dict(
            {
                "browser_type": ToEnumTrafaret(BrowserType),
                "width": trf.Int,
                "height": trf.Int,
            }
        )
    )

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return [DesktopBrowserInfo(**dct) for dct in sanitized]


class IosDeviceInfoTrafaret(trf.Trafaret):
    scheme = trf.List(
        trf.Dict(
            {
                "device_name": ToEnumTrafaret(IosDeviceName),
                trf.Key("screen_orientation", optional=True): ToEnumTrafaret(
                    ScreenOrientation
                ),
                trf.Key("ios_version", optional=True): ToEnumTrafaret(IosVersion),
            }
        )
    )

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return [IosDeviceInfo(**dct) for dct in sanitized]


class ChromeEmulationInfoTrafaret(trf.Trafaret):
    scheme = trf.List(
        trf.Dict(
            {
                "device_name": ToEnumTrafaret(DeviceName),
                trf.Key("screen_orientation", optional=True): ToEnumTrafaret(
                    ScreenOrientation
                ),
            }
        )
    )

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return [ChromeEmulationInfo(**dct) for dct in sanitized]


class RunnerOptionsTrafaret(trf.Trafaret):
    scheme = trf.Dict({trf.Key("test_concurrency"): trf.Int})

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return RunnerOptions().test_concurrency(sanitized["test_concurrency"])


class ProxyTrafaret(trf.Trafaret):
    scheme = trf.Dict({trf.Key("url") >> "host_or_url": trf.URL}) | trf.Dict(
        {
            trf.Key("host") >> "host_or_url": trf.String,
            "port": trf.Int,
            "username": trf.String,
            "password": trf.String,
            "scheme": trf.String,
        },
    )

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        return ProxySettings(host_or_url=sanitized.pop("host_or_url"), **sanitized)


class BrowsersTrafaret(trf.Trafaret):
    scheme = trf.Dict(
        {
            trf.Key("desktop", optional=True): DesktopBrowserInfoTrafaret,
            trf.Key("ios", optional=True): IosDeviceInfoTrafaret,
            trf.Key("chrome_emulation", optional=True): ChromeEmulationInfoTrafaret,
        }
    )

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        desktop = sanitized.pop("desktop", [])
        ios = sanitized.pop("ios", [])
        chrome_emulation = sanitized.pop("chrome_emulation", [])
        return list(itertools.chain(desktop, ios, chrome_emulation))


class ConfigurationTrafaret(trf.Trafaret):  # typedef
    shared_scheme = trf.Dict(
        {
            trf.Key("server_url", optional=True): trf.URL,
            trf.Key("batch", optional=True): BatchInfoTrafaret,
            trf.Key("proxy", optional=True): ProxyTrafaret,
            trf.Key("app_name", optional=True): trf.String,
            trf.Key("api_key", optional=True): trf.String,
            trf.Key("branch_name", optional=True): trf.String,
            trf.Key("parent_branch_name", optional=True): trf.String,
            trf.Key("baseline_branch_name", optional=True): trf.String,
            trf.Key("baseline_env_name", optional=True): trf.String,
            trf.Key("save_diffs", optional=True): trf.Bool,
            trf.Key("app_name", optional=True): trf.String,
            trf.Key("viewport_size", optional=True): ViewPortTrafaret,
            trf.Key("match_timeout", optional=True): trf.Int,
            trf.Key("save_new_tests", optional=True): trf.Bool,
            trf.Key("save_failed_tests", optional=True): trf.Bool,
            trf.Key("properties", optional=True): trf.List(
                trf.Dict(name=trf.String, value=trf.String)
            ),
        },
    )
    selenium_scheme = shared_scheme + trf.Dict(
        {
            trf.Key("force_full_page_screenshot", optional=True): trf.Bool,
            trf.Key("wait_before_screenshots", optional=True): trf.Int,
            trf.Key("stitch_mode", optional=True): ToEnumTrafaret(RobotStitchMode),
            trf.Key("hide_scrollbars", optional=True): trf.Bool,
            trf.Key("hide_caret", optional=True): trf.Bool,
        },
    )
    selenium_ufg_scheme = shared_scheme + trf.Dict(
        {
            trf.Key("runner_options", optional=True): RunnerOptionsTrafaret,
            trf.Key("visual_grid_options", optional=True): VisualGridOptionsTrafaret,
            trf.Key("disable_browser_fetching", optional=True): trf.Bool,
            trf.Key("enable_cross_origin_rendering", optional=True): trf.Bool,
            trf.Key("dont_use_cookies", optional=True): trf.Bool,
            trf.Key("layout_breakpoints", optional=True): trf.Bool | trf.List(trf.Int),
            trf.Key("browsers", optional=True) >> "_browsers_info": BrowsersTrafaret,
        },
    )
    appium_scheme = shared_scheme + trf.Dict(
        {
            trf.Key("is_simulator", optional=True): trf.Bool,
        },
    )
    scheme = shared_scheme + trf.Dict(
        {
            trf.Key("web", optional=True): shared_scheme + selenium_scheme,
            trf.Key("web_ufg", optional=True): shared_scheme + selenium_ufg_scheme,
            trf.Key("mobile_native", optional=True): shared_scheme + appium_scheme,
        },
    )

    def __init__(self, selected_sdk, add_to_config):
        # type: (SelectedRunner, RobotConfiguration) -> None
        self._selected_runner = selected_sdk
        self._exists_configuration = add_to_config

    def check_and_return(self, value, context=None):
        sanitized = self.scheme.check(value, context)
        selected_sdk_conf = sanitized[self._selected_runner.value]

        combined_raw_config = sanitized.copy()
        # we need only shared data here  TODO: make it nicer
        combined_raw_config.pop(SelectedRunner.web.value)
        combined_raw_config.pop(SelectedRunner.web_ufg.value)
        combined_raw_config.pop(SelectedRunner.mobile_native.value)

        # add config for selected runner
        combined_raw_config.update(selected_sdk_conf)
        for key, val in combined_raw_config.items():
            setattr(self._exists_configuration, key, val)
        return self._exists_configuration


def try_parse_runner(runner):
    # type: (Text) -> SelectedRunner
    try:
        return SelectedRunner(runner)
    except ValueError as e:
        raise_from(
            EyesLibraryValueError(
                "Incorrect value for `runner`: `{val}`. "
                "\n\tPossible variants:"
                "\n\t\t{possible_vals}".format(
                    val=runner,
                    possible_vals=", ".join(e.name for e in SelectedRunner),
                )
            ),
            e,
        )


def try_parse_configuration(
    config_path, selected_runner, origin_configuration, suite_path
):
    # type: (Text, SelectedRunner, RobotConfiguration, Text) -> RobotConfiguration
    if not os.path.isabs(config_path):
        # if config path is not absolute that count test suite directory as root
        config_path = os.path.join(suite_path, config_path)

    if not os.path.exists(config_path):
        raise EyesLibraryConfigError(
            "The configuration file was not found in the following directory: {}\n"
            "You could initialize config file with command: \n\t"
            "`python -m EyesLibrary init-config`".format(config_path)
        )

    with open(config_path, "r") as f:
        raw_config = yaml.safe_load(f.read())

    # It's better to raise here library error but for some reason raise_from
    # suppress original error message
    return ConfigurationTrafaret(selected_runner, origin_configuration).check(
        raw_config
    )



import os
import traceback
import typing
from typing import TYPE_CHECKING, Dict, Union

from robot.api import logger as robot_logger
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from robotlibcore import DynamicCore
from six import raise_from
from six import string_types as basestring

from applitools.common import BatchInfo
from applitools.common.utils import argument_guard
from applitools.selenium import ClassicRunner, Eyes, VisualGridRunner
from applitools.selenium.eyes import EyesRunner

from .__version__ import __version__
from .config import RobotConfiguration
from .config_parser import (
    ConfigurationTrafaret,
    SelectedRunner,
    try_parse_configuration,
    try_parse_runner,
)
from .errors import EyesLibraryConfigError, EyesLibraryError
from .eyes_cache import EyesCache
from .keywords import (
    CheckKeywords,
    CheckSettingsKeywords,
    ConfigurationKeywords,
    SessionKeywords,
    TargetKeywords,
)
from .keywords.session import RunnerKeywords
from .library_listener import LibraryListener
from .locator_converter import LocatorConverter

if TYPE_CHECKING:
    from typing import TYPE_CHECKING, Optional, Text

    from applitools.common.utils.custom_types import AnyWebDriver


def get_suite_path():
    suite_source = BuiltIn().get_variable_value("${SUITE_SOURCE}")
    if os.path.isdir(suite_source):
        return suite_source
    return os.path.dirname(suite_source)


def is_test_run():
    try:
        BuiltIn()._get_context()
        return True
    except RobotNotRunningError:
        # run without test suite, probably `libdoc` generation
        return False


def validate_config(configuration):
    if configuration.api_key is None:
        raise EyesLibraryConfigError(
            "API key not set! Log in to https://applitools.com to obtain "
            "your API Key and set it to `applitools.yaml` or `APPLITOOLS_API_KEY`."
        )


class EyesLibrary(DynamicCore):
    """
    EyesLibrary is a visual verification library for [http://robotframework.org/|Robot Framework] that uses
    [https://applitools.com/docs/api/eyes-sdk/index-gen/classindex-selenium-python_sdk4.html|Applitools Eyes SDK Python] and
    [http://robotframework.org/SeleniumLibrary/SeleniumLibrary.html|SeleniumLibrary] /
    [http://serhatbolsu.github.io/robotframework-appiumlibrary/AppiumLibrary.html|AppiumLibrary].

    = Table of contents =
    - `Preconditions`
    - `Configuration`
    - `Writing tests`
    - `Importing`
    - `Keywords`

    = Preconditions =
    To run tests, you need to have the Applitools API key. If you have an Applitools account,
    you could fetch it from [https://eyes.applitools.com/app/admin/api-keys|dashboard],
    or you could create the [https://applitools.com/sign-up/|free account].
    You may want to read [https://applitools.com/docs|Applitools documentation] to understand better how Eyes works.

    = Configuration =
    EyesLibrary stores configuration file inside `applitools.yaml` config file.
     Detailed info you can find in [https://applitools.com/docs/api/robot/robot-configuration-file.html|configuration docs]

    The general options provide defaults for all run types. The other sections are for specific types of run and define their own configuration values. By default, they inherit the general options, but they can override any configuration if necessary.

    Before running tests, you must initialize `applitools.yaml` configuration script in root of test suite:
        | python -m EyesLibrary init-config |

    Add your *API KEY* `applitools.yaml` or in *APPLITOOLS_API_KEY* env variable and import EyesLibrary into your Robot test suite:
        | Library | EyesLibrary | runner=web | config=path/to/applitools.yaml |

    = Writing tests =
    When writing the tests, the following structure must be adopted:

    1. *Eyes Open*

    A browser or application must be running when opening the session.
    To open a browser/application, consult the documentation for [http://robotframework.org/SeleniumLibrary/SeleniumLibrary.html|SeleniumLibrary]
    and/or [http://serhatbolsu.github.io/robotframework-appiumlibrary/AppiumLibrary.html|AppiumLibrary].

    Afterwards, the session may be opened. See `Eyes Open`.

    2. *Visual Checks*

    Between opening and closing the session, you can run your visual checks.

    See `Eyes Check Window`, `Eyes Check Region By Element`, `Eyes Check Region By Selector`, `Eyes Check Region By Coordinates`,
    `Eyes Check Frame By Element`, `Eyes Check Frame By Name`, `Eyes Check Frame By Index` and `Eyes Check Frame By Selector`.

    You can also verify if there's an open session with `Is Eyes Open`.

    3. *Eyes Close Async*

    See `Eyes Close Async`.

    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_VERSION = __version__
    _eyes_runner = None  # type: Optional[VisualGridRunner, ClassicRunner]
    driver = None  # type: Optional[AnyWebDriver]
    _selected_runner = None  # type: Optional[SelectedRunner]
    supported_library_names_by_runner = {
        SelectedRunner.web: ("SeleniumLibrary", "AppiumLibrary"),
        SelectedRunner.web_ufg: ("SeleniumLibrary",),
        SelectedRunner.mobile_native: ("AppiumLibrary",),
    }

    def __init__(
        self,
        runner=None,  # type: Text
        config=None,  # type: Text
        run_on_failure="Eyes Abort Async",
    ):
        # type: (...) -> None
        """
        Initialize the EyesLibrary
            | =Arguments=      | =Description=  |
            | runner           | Specify one of `web`, `web_ufg`, or `mobile_native` runners (by default `web`)  |
            | config           | Path to applitools.yaml (if no specify, trying to find it in test suite dir)  |
            | run_on_failure   | Specify keyword to run in case of failure (By default `Eyes Abort Async`)  |

        """
        # skip loading of dynamic libraries during doc generation
        if config is None:
            # try to find `applitools.yaml` in test directory
            robot_logger.warn(
                "No `config` set. Trying to find `applitools.yaml` in current path"
            )
            config = "applitools.yaml"

        if runner is None:
            runner = SelectedRunner.web
            robot_logger.warn("No `runner` set. Using `web` runner.")

        robot_logger.console(
            "Running test suite with `{}` runner and `{}` config".format(runner, config)
        )

        self.run_on_failure_keyword = run_on_failure

        self._running_on_failure_keyword = False
        self._eyes_registry = EyesCache()
        self._batch_registry = {}  # type: Dict[Text, BatchInfo]
        self._running_keyword = None
        self._configuration = None

        self._selected_runner = try_parse_runner(runner)

        if is_test_run():
            self.current_library = self._try_get_library(self._selected_runner)
            suite_path = get_suite_path()
            self._configuration = try_parse_configuration(
                config, self._selected_runner, RobotConfiguration(), suite_path
            )
            validate_config(self._configuration)
            self.ROBOT_LIBRARY_LISTENER = LibraryListener(self)
            self._locator_converter = LocatorConverter(self)
        else:
            # hide objects that uses dynamic loading for generation of documentation
            self.current_library = None

        keywords = [
            RunnerKeywords(self),
            SessionKeywords(self),
            CheckKeywords(self),
            TargetKeywords(self),
            CheckSettingsKeywords(self),
            ConfigurationKeywords(self),
        ]

        DynamicCore.__init__(self, keywords)

    @property
    def eyes_runner(self):
        # type: () -> EyesRunner
        return self._eyes_runner

    @eyes_runner.setter
    def eyes_runner(self, runner):
        # type: (EyesRunner) -> None
        argument_guard.is_a(runner, (ClassicRunner, VisualGridRunner))
        self._eyes_runner = runner

    def clean_eyes_runner(self):
        self._eyes_runner = None

    def _try_get_library(self, runner):
        # type: (SelectedRunner) -> typing.ForwardRef
        """Check if supported library was loaded"""
        library_names = self.supported_library_names_by_runner[runner]
        supported_library = None
        failed_to_import = []
        for library_name in library_names:
            try:
                supported_library = BuiltIn().get_library_instance(name=library_name)
                robot_logger.console(
                    "Using library `{}` as backend".format(library_name)
                )
                break
            except RuntimeError:
                failed_to_import.append(library_name)

        if supported_library is None:
            raise EyesLibraryError(
                "Failed to find libraries {} for runner: {} in your test suite".format(
                    library_names, runner
                )
            )
        return supported_library

    def run_keyword(self, name, *args, **kwargs):
        try:
            return DynamicCore.run_keyword(self, name, *args, **kwargs)
        except Exception as e:
            trb_text = traceback.format_exc()
            self.failure_occurred(e, trb_text)

    def failure_occurred(self, origin_exc, trb_text):
        """Method that is executed when a keyword fails."""
        if self._running_on_failure_keyword or not self.run_on_failure_keyword:
            return
        try:
            self._running_on_failure_keyword = True
            if self.run_on_failure_keyword.lower() == "eyes abort":
                self.current_eyes.abort_async()
            else:
                BuiltIn().run_keyword(self.run_on_failure_keyword)
        except Exception as err:
            robot_logger.warn(
                "Keyword '{}' could not be run on failure: {}".format(
                    self.run_on_failure_keyword, err
                )
            )
        finally:
            self._running_on_failure_keyword = False
            raise_from(
                EyesLibraryError("Failed to run EyesLibrary\n{}".format(trb_text)),
                origin_exc,
            )

    def register_eyes(self, eyes, alias):
        """Add's a `Eyes` to the library EyesCache."""
        return self._eyes_registry.register(eyes, alias)

    def register_or_get_batch(self, batch):
        # type: (Union[basestring, BatchInfo]) -> BatchInfo
        if isinstance(batch, basestring):
            return self._batch_registry.setdefault(batch, BatchInfo(batch))
        elif isinstance(batch, BatchInfo):
            return self._batch_registry.setdefault(batch.id, batch)
        else:
            raise ValueError("Not supported `batch` value")

    @property
    def current_eyes(self):
        # type: () -> Eyes
        if not self._eyes_registry.current:
            raise RuntimeError("No opened Eyes.")
        return self._eyes_registry.current

    @property
    def selected_runner(self):
        # type: () -> SelectedRunner
        if self._selected_runner is None:
            raise RuntimeError("No runner have been selected yet.")
        return self._selected_runner

    @property
    def configure(self):
        return self._configuration



class EyesLibraryError(RuntimeError):
    ROBOT_EXIT_ON_FAILURE = True


class EyesLibraryValueError(ValueError):
    pass


class EyesLibraryConfigError(EyesLibraryValueError):
    ROBOT_EXIT_ON_FAILURE = True
    pass



from .check import CheckKeywords  # noqa
from .check_settings import CheckSettingsKeywords  # noqa
from .config import ConfigurationKeywords  # noqa
from .session import SessionKeywords  # noqa
from .target import TargetKeywords  # noqa



from __future__ import absolute_import, unicode_literals

from typing import TYPE_CHECKING, Optional, Text

from appium.webdriver import WebElement as AppiumWebElement
from robot.api.deco import keyword as original_keyword
from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement

from applitools.common import AccessibilityRegionType, MatchLevel, VisualGridOption
from applitools.selenium.fluent import SeleniumCheckSettings

from ..base import LibraryComponent
from ..utils import get_enum_by_name, is_webelement_guard, parse_region
from .keyword_tags import CHECK_SETTING, UFG_RELATED

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import AnyWebElement

    from ..custom_types import Locator


def keyword(name=None, tags=(), types=()):
    """Keyword with predefined CHECK_SETTING tag"""
    tags = tags + (CHECK_SETTING,)
    return original_keyword(name, tags, types)


def new_or_cur_check_settings(check_settings):
    # type: (Optional[SeleniumCheckSettings])->SeleniumCheckSettings
    if check_settings is None:
        return SeleniumCheckSettings()
    return check_settings


class IgnoreCheckSettingsKeyword(object):
    @keyword("Ignore Region By Coordinates", types=(str,))
    def ignore_region_by_coordinates(
        self,
        region,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object that ignores the region specified in the argument.

            | =Arguments=   | =Description=                                                       |
            | Region          | *Mandatory* - The region to ignore in format [left top width height] ,e.g. [100 200 300 300]  |

        *Example:*
            | Ignore Region By Coordinates           | [10 20 100 100]  |
        """
        return new_or_cur_check_settings(check_settings).ignore(parse_region(region))

    @keyword(
        "Ignore Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
    )
    def ignore_region_by_element(
        self,
        element,  # type: AnyWebElement
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object that ignores the region specified in the argument.

            | =Arguments=   | =Description=                                           |
            | Element       | *Mandatory* - The element to ignore                     |

        *Example:*
            | ${element}=   | Get Webelement                | //*[@id="logo"]   |
            | ${target}=    | Ignore Region By Element      | ${element}        |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).ignore(element)

    @keyword("Ignore Region By Selector", types=(str,))
    def ignore_region_by_selector(
        self,
        selector,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object that ignores the region specified in the argument.

            | =Arguments=   | =Description=                                               |
            | Selector       | *Mandatory* - The selector for element to ignore. Selenium/Appium formats are supported. |

        *Example:*
            | Ignore Region By Selector      | css:div         |
        """
        return new_or_cur_check_settings(check_settings).ignore(
            *self.from_locators_to_supported_form(selector)
        )


class LayoutCheckSettingsKeyword(object):
    @keyword("Layout Region By Coordinates", types=(str,))
    def layout_region_by_coordinates(
        self,
        region,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with layout region specified in the argument.

            | =Arguments=   | =Description=                                                       |
            | Region        | *Mandatory* - The layout region in format [left top width height] ,e.g. [100 200 300 300]  |

        *Example:*
            | Leyout Region By Coordinates           | [10 20 100 100]  |
        """
        return new_or_cur_check_settings(check_settings).layout(parse_region(region))

    @keyword(
        "Layout Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
    )
    def layout_region_by_element(
        self,
        element,  # type: AnyWebElement
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with layout region specified in the argument.

            | =Arguments=   | =Description=                                               |
            | Element       | *Mandatory* - The element with layout region e.g. [100 200 300 300]  |

        *Example:*
            | ${element}=   | Get Webelement                | //*[@id="logo"]   |
            | ${target}=    | Layout Region By Element      | ${element}        |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).layout(element)

    @keyword("Layout Region By Selector", types=(str,))
    def layout_region_by_selector(
        self,
        selector,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with layout region specified in the argument.

            | =Arguments=   | =Description=                                               |
            | Selector      | *Mandatory* - The selector for element for layout region. Selenium/Appium formats are supported. |

        *Example:*
            | Layout Region By Selector      | css:div         |
        """
        return new_or_cur_check_settings(check_settings).layout(
            *self.from_locators_to_supported_form(selector)
        )


class ContentCheckSettingsKeyword(object):
    @keyword("Content Region By Coordinates", types=(str,))
    def content_region_by_coordinates(
        self,
        region,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with content region specified in the argument.

            | =Arguments=   | =Description=                                                       |
            | Region        | *Mandatory* - The content region in format [left top width height] ,e.g. [100 200 300 300]  |

        *Example:*
            | Content Region By Coordinates           | [10 20 100 100]  |
        """
        return new_or_cur_check_settings(check_settings).content(parse_region(region))

    @keyword(
        "Content Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
    )
    def content_region_by_element(
        self,
        element,  # type: AnyWebElement
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with content region specified in the argument.

            | =Arguments=   | =Description=                                       |
            | Element       | *Mandatory* - The element to become content region  |

        *Example:*
            | ${element}=   | Get Webelement                | //*[@id="logo"]   |
            | ${target}=    | Content Region By Element     | ${element}        |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).content(element)

    @keyword("Content Region By Selector", types=(str,))
    def content_region_by_selector(
        self,
        selector,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with content region specified in the argument.

            | =Arguments=   | =Description=                                               |
            | Selector      | *Mandatory* - The selector for element for content region. Selenium/Appium formats are supported. |

        *Example:*
            | Content Region By Selector      | css:div         |
        """
        return new_or_cur_check_settings(check_settings).content(
            *self.from_locators_to_supported_form(selector)
        )


class StrictCheckSettingsKeywords(object):
    @keyword("Strict Region By Coordinates", types=(str,))
    def strict_region_by_coordinates(
        self,
        region,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with content region specified in the argument.

            | =Arguments=   | =Description=                                                       |
            | Region        | *Mandatory* - The strict region in format [left top width height] ,e.g. [100 200 300 300]  |

        *Example:*
            | Strict Region By Coordinates           | [10 20 100 100]  |
        """
        return new_or_cur_check_settings(check_settings).strict(parse_region(region))

    @keyword(
        "Strict Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
    )
    def strict_region_by_element(
        self,
        element,  # type: AnyWebElement
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with strict region specified in the argument.

            | =Arguments=   | =Description=                                         |
            | Element       | *Mandatory* - The element to become strict region     |

        *Example:*
            | ${element}=   | Get Webelement                | //*[@id="logo"]   |
            | ${target}=    | Strict Region By Element      | ${element}        |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).strict(element)

    @keyword("Strict Region By Selector", types=(str,))
    def strict_region_by_selector(
        self,
        selector,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with strict region specified in the argument.

            | =Arguments=   | =Description=                                               |
            | Selector      | *Mandatory* - The selector for element for strict region. Selenium/Appium formats are supported. |

        *Example:*
            | Eyes Check                         |          |
            | ...  Strict Region By Selector     | css:div  |
        """
        return new_or_cur_check_settings(check_settings).strict(
            *self.from_locators_to_supported_form(selector)
        )


class FloatingCheckSettingsKeywords(object):
    @keyword(
        "Floating Region With Max Offset By Coordinates",
        types=(int, str),
    )
    def floating_region_with_max_offset_by_coordinates(
        self,
        max_offset,  # type: int
        region,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with floating region specified in the arguments.

            | =Arguments=  | =Description=                                                                                                 |
            | Max Offset   | *Mandatory* - The maximum amount that the region can shift in any direction and still be considered matching. |
            | Region       | *Mandatory* - The floating region e.g. [100 200 300 300]                                                       |

        *Example:*
            | Eyes Check                                            |     |                   |
            | ...   Floating Region With Max Offset By Coordinates  |  5  |  [10 20 100 100]  |
        """
        return new_or_cur_check_settings(check_settings).floating(
            max_offset, parse_region(region)
        )

    @keyword(
        "Floating Region With Max Offset By Element",
        types={"max_offset": int, "element": (SeleniumWebElement, AppiumWebElement)},
    )
    def floating_region_with_max_offset_by_element(
        self,
        max_offset,  # type: int
        element,  # type: AnyWebElement
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with floating region specified in the arguments.

            | =Arguments=  | =Description=                                                                                                 |
            | Max Offset   | *Mandatory* - The maximum amount that the region can shift in any direction and still be considered matching. |
            | Element       | *Mandatory* - The element to become floating region

        *Example:*
            | ${element}=   | Get Webelement                        |  //*[@id="logo"]  |              |
            | Eyes Check                                            |                   |              |
            | ...   Floating Region With Max Offset By Element      |       5           |  ${element}  |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).floating(max_offset, element)

    @keyword(
        "Floating Region With Max Offset By Selector",
        types=(str, str),
    )
    def floating_region_with_max_offset_by_selector(
        self,
        max_offset,  # type: int
        selector,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with floating region specified in the arguments.

            | =Arguments=  | =Description=                                                                                                 |
            | Max Offset   | *Mandatory* - The maximum amount that the region can shift in any direction and still be considered matching. |
            | Selector     | *Mandatory* - The selector to become floating region e.g. //*[@id="logo"]                                     |

        *Example:*
            | Eyes Check                                            |                   |                    |
            | ...   Floating Region With Max Offset By Selector     |       5           |  //*[@id="logo"]   |
        """
        return new_or_cur_check_settings(check_settings).floating(
            max_offset, self.from_locator_to_supported_form(selector)
        )

    @keyword(
        "Floating Region By Coordinates",
        types=(str, int, int, int, int),
    )
    def floating_region_by_coordinates(
        self,
        region,  # type: Text
        max_up_offset=0,  # type: int
        max_down_offset=0,  # type: int
        max_left_offset=0,  # type: int
        max_right_offset=0,  # type: int
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with floating region specified in the arguments.

            | =Arguments=       | =Description=                                                                                              |
            | Region            | *Mandatory* - The floating region e.g. [100 200 300 300]                                                        |
            | Max Up Offset     | The maximum amount that the region can shift upwards and still be considered matching.       |
            | Max Down Offset   | The maximum amount that the region can shift downwards and still be considered matching.     |
            | Max Left Offset   | The maximum amount that the region can shift to the left and still be considered matching.   |
            | Max Right Offset  | The maximum amount that the region can shift to the right and still be considered matching.  |

        *Example:*
            | Eyes Check Window                       |                   |     |     |    |     |
            | ...     Floating Region By Coordinates  |  [10 20 100 100]  |  5  |  5  | 5  |  5  |
            | ...     Floating Region By Coordinates  |  [10 20 100 100]  |  max_right_offset=5  |   |   |    |
        """
        return new_or_cur_check_settings(check_settings).floating(
            parse_region(region),
            max_up_offset,
            max_down_offset,
            max_left_offset,
            max_right_offset,
        )

    @keyword(
        "Floating Region By Element",
        types={
            "element": (SeleniumWebElement, AppiumWebElement),
            "max_up_offset": int,
            "max_down_offset": int,
            "max_left_offset": int,
            "max_right_offset": int,
        },
    )
    def floating_region_by_element(
        self,
        element,  # type: AnyWebElement
        max_up_offset=0,  # type: int
        max_down_offset=0,  # type: int
        max_left_offset=0,  # type: int
        max_right_offset=0,  # type: int
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with floating region specified in the arguments.

            | =Arguments=      | =Description=                                                                                                       |
            | Element          | *Mandatory* - The WebElement to set as float region.                                                  |
            | Max Up Offset    | The maximum amount that the region can shift upwards and still be considered matching.             |
            | Max Down Offset  | The maximum amount that the region can shift downwards and still be considered matching.         |
            | Max Left Offset  | The maximum amount that the region can shift to the left and still be considered matching.       |
            | Max Right Offset | The maximum amount that the region can shift to the right and still be considered matching.     |

        *Example:*
            | ${element}=                           |  Get Webelement  |  //*[@id="logo"]  |  |  |  |
            | Eyes Check Window                     |               |     |     |     |     |
            | ...   Floating Region By Element      |  ${element}   |  5  |  5  |  5  |  5  |
            | ...   Floating Region By Element      |  ${element}   |  max_left_offset=5 |   |   |    |

        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).floating(
            element,
            max_up_offset,
            max_down_offset,
            max_left_offset,
            max_right_offset,
        )

    @keyword(
        "Floating Region By Selector",
        types=(str, int, int, int, int),
    )
    def floating_region_by_selector(
        self,
        selector,  # type: Text
        max_up_offset=0,  # type: int
        max_down_offset=0,  # type: int
        max_left_offset=0,  # type: int
        max_right_offset=0,  # type: int
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with floating region specified in the arguments.

            | =Arguments=       | =Description=                                                                                |
            | Selector          | The selector to become floating region e.g. //*[@id="logo"]                                  |
            | Max Up Offset     | The maximum amount that the region can shift upwards and still be considered matching.       |
            | Max Down Offset   | The maximum amount that the region can shift downwards and still be considered matching.     |
            | Max Left Offset   | The maximum amount that the region can shift to the left and still be considered matching.   |
            | Max Right Offset  | The maximum amount that the region can shift to the right and still be considered matching.  |

        *Example:*
            | Eyes Check Window                       |                  |     |     |     |      |
            | ...     Floating Region By Selector     |  //*[@id="logo"] |  5  |  5  |  5  |  5   |
            | ...     Floating Region By Selector     |  //*[@id="logo"] |  max_left_offset=5 |   |   |    |
        """
        return new_or_cur_check_settings(check_settings).floating(
            self.from_locator_to_supported_form(selector),
            max_up_offset,
            max_down_offset,
            max_left_offset,
            max_right_offset,
        )


class AccessibilityCheckSettingsKeywords(object):
    @keyword("Accessibility Region By Selector", types=(str, str))
    def accessibility_region_by_selector(
        self,
        selector,  # type: Text
        type,  # type: AccessibilityRegionType
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with accessibility region specified in the arguments.

            | =Arguments=   |   =Description=                                          |
            | Selector      |  *Mandatory* - The selector for element for accessibility region. Selenium/Appium formats are supported. |
            | Type          | *Mandatory* - Type of AccessibilityRegion. (`IgnoreContrast`, `RegularText`, `LargeText`, `BoldText`, `GraphicalObject`)    |

        *Example:*
            | Eyes Check                                |              |                  |
            | ...     Accessibility Region By Selector  |  //selector  |  GraphicalObject |
        """
        return new_or_cur_check_settings(check_settings).accessibility(
            *self.from_locators_to_supported_form(selector),
            type=AccessibilityRegionType(type)
        )

    @keyword(
        "Accessibility Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement), "type": str},
    )
    def accessibility_region_by_element(
        self,
        element,  # type: Text
        type,  # type: AccessibilityRegionType
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with accessibility region specified in the arguments.

            | =Arguments=      | =Description=                                                            |
            | Element          | *Mandatory* - The accessibility region e.g. [100 200 300 300]                                                             |
            | Type             | *Mandatory* - Type of AccessibilityRegion. (`IgnoreContrast`, `RegularText`, `LargeText`, `BoldText`, `GraphicalObject`)  |

        *Example:*
            | Eyes Check                                   |              |                   |
            | ...     Accessibility Region By Coordinates  |  ${element}  |  GraphicalObject  |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).accessibility(
            element, type=AccessibilityRegionType(type)
        )

    @keyword("Accessibility Region By Coordinates", types=(str, str))
    def accessibility_region_by_coordinates(
        self,
        region,  # type: Text
        type,  # type: AccessibilityRegionType
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with accessibility region specified in the arguments.

            | =Arguments=      | =Description=                                                            |
            | Region           | *Mandatory* - The accessibility region e.g. [100 200 300 300]                                                        |
            | Type             | *Mandatory* - Type of AccessibilityRegion. (`IgnoreContrast`, `RegularText`, `LargeText`, `BoldText`, `GraphicalObject`)    |

        *Example:*
            | Eyes Check                                   |                 |                   |
            | ...     Accessibility Region By Coordinates  |  [10 20 30 40]  |  GraphicalObject  |
        """
        return new_or_cur_check_settings(check_settings).accessibility(
            region=parse_region(region), type=AccessibilityRegionType(type)
        )


class UFGCheckSettingsKeywords(object):
    @keyword("Visual Grid Option", types=(str, str), tags=(UFG_RELATED,))
    def visual_grid_option(
        self,
        name,  # type: Text
        value,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with VisualGridOption specified in the arguments.

            | =Arguments=   | =Description=                              |                                                           |
            | name          | *Mandatory* - The VisualGridOption name.   |                                                     |                                                         |
            | value         | *Mandatory* - The VisualGridOption value.  |

        *Example:*
            | Eyes Check                  |
            | ...     Visual Grid Option  |  key name     value   |
            | ...     Visual Grid Option  |  key name2    value   |
        """
        return new_or_cur_check_settings(check_settings).visual_grid_options(
            VisualGridOption(name, value)
        )

    @keyword("Disable Browser Fetching", tags=(UFG_RELATED,))
    def disable_browser_fetching(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with disable_browser_fetching specified in the arguments.

        *Example:*
            | Eyes Check                        |           |
            | ...     Disable Browser Fetching  |           |
        """
        return new_or_cur_check_settings(check_settings).disable_browser_fetching(True)

    @keyword("Enable Layout Breakpoints", tags=(UFG_RELATED,))
    def enable_layout_breakpoints(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with enabled layout_breakpoints specified in the arguments.

        *Example:*
            | Eyes Check                         |          |
            | ...     Enable Layout Breakpoints  |          |
        """
        return new_or_cur_check_settings(check_settings).layout_breakpoints(True)

    @keyword("Layout Breakpoints", types=(str,), tags=(UFG_RELATED,))
    def layout_breakpoints(
        self,
        breakpoints,  # type:str
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified layout_breakpoints in the arguments.

            | =Arguments=     | =Description=                             |
            | breakpoints     | Specify layout breakpoint, e.g. 25 56 89  |

        *Example:*
            | Eyes Check  |  Target Window  |  Layout Breakpoints  |  23 45 67 89 |

        """
        breakpoints = [int(b) for b in breakpoints.split(" ")]
        return new_or_cur_check_settings(check_settings).layout_breakpoints(
            *breakpoints
        )

    @keyword("Before Render Screenshot Hook", types=(str,), tags=(UFG_RELATED,))
    def before_render_screenshot_hook(
        self,
        hook,  # type:str
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        return new_or_cur_check_settings(check_settings).before_render_screenshot_hook(
            hook
        )


class CheckSettingsKeywords(
    LibraryComponent,
    IgnoreCheckSettingsKeyword,
    LayoutCheckSettingsKeyword,
    ContentCheckSettingsKeyword,
    StrictCheckSettingsKeywords,
    FloatingCheckSettingsKeywords,
    AccessibilityCheckSettingsKeywords,
    UFGCheckSettingsKeywords,
):
    @keyword("Scroll Root Element By Selector", types=(str,))
    def scroll_root_element_by_selector(
        self,
        selector,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Scroll Root Element by selector

            |  =Arguments=  | =Description=                                          |
            | Selector       | *Mandatory* - The selector of scroll root element     |

        *Example:*
            |  Scroll Root Element By Element  |  ${element}  |
        """
        return new_or_cur_check_settings(check_settings).scroll_root_element(
            self.from_locator_to_supported_form(selector)
        )

    @keyword(
        "Scroll Root Element By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
    )
    def scroll_root_element_by_element(
        self,
        element,  # type: Locator
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Scroll Root Element

            |  =Arguments=  | =Description=                                          |
            | Element       | *Mandatory* - The scroll root element                  |

        *Example:*
            |  Scroll Root Element By Element  |  ${element}  |
        """
        is_webelement_guard(element)
        return new_or_cur_check_settings(check_settings).scroll_root_element(element)

    @keyword("Variation Group Id", types=(str,))
    def variation_group_id(
        self,
        variation_group_id,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified variant group id.

            | =Arguments=  | =Description=                                                      |
            | variation_group_id  | will be associated with all of the test result steps that result from executing this checkpoint |

        *Example:*
            | Eyes Check   |  Target Window  |  Variation Group Id  |  variation1 |
        """
        return new_or_cur_check_settings(check_settings).variation_group_id(
            variation_group_id
        )

    @keyword("Match Level", types=(str,))
    def match_level(
        self,
        match_level,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified match level.

            | =Arguments=  | =Description=                                                      |
            | match_level  | Specify the match level that should be used for the target area (NONE LAYOUT CONTENT STRICT EXACT) |

        *Example:*
            | Eyes Check   |  Target Window  |  Match Level  STRICT |
        """
        match_level = get_enum_by_name(match_level, MatchLevel)
        return new_or_cur_check_settings(check_settings).match_level(match_level)

    @keyword("Enable Patterns")
    def enable_patterns(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified enable patterns.

        *Example:*
            | Eyes Check   |  Target Window  |  Enable Patterns  |
        """
        return new_or_cur_check_settings(check_settings).enable_patterns(True)

    @keyword("Ignore Displacements")
    def ignore_displacements(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified ignore displacements.

        *Example:*
            | Eyes Check   |  Target Window  |  Ignore Displacements |
        """
        return new_or_cur_check_settings(check_settings).ignore_displacements(True)

    @keyword("Ignore Caret")
    def ignore_caret(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified ignore caret.

        *Example:*
            | Eyes Check   |  Target Window  |  Ignore Caret |
        """
        return new_or_cur_check_settings(check_settings).ignore_caret(True)

    @keyword("Fully")
    def fully(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified fully.

        *Example:*
            | Eyes Check   |  Target Window  |  Fully |
        """
        return new_or_cur_check_settings(check_settings).fully(True)

    @keyword("With Name", types=(str,))
    def with_name(
        self,
        name,  # type: Text
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified tag from argument.

            | =Arguments=  | =Description=                                                                |
            | name         | Specifies a tag for this target (instead of a parameter to the Eyes Check )  |

        *Example:*
            | Eyes Check         |    Target Window        |
            | ...     With Name  |   User Check Step name  |
        """
        return new_or_cur_check_settings(check_settings).with_name(name)

    @keyword("Timeout", types=(int,))
    def timeout(
        self,
        timeout,  # type: int
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified maximum amount of time that Eyes should retry capturing the window content if there are mismatches.

            | =Arguments=  | =Description=                                       |
            | timeout      | Specify timeout in milliseconds, e.g. 1000 (1 sec)  |

        *Example:*
            | Eyes Check         |    Target Window    |
            | ...     Timeout    |   3000              |
        """
        return new_or_cur_check_settings(check_settings).timeout(timeout)

    @keyword("Use Dom")
    def use_dom(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified use dom.

        *Example:*
            | Eyes Check   |    Target Window    |  Use Dom   |
        """
        return new_or_cur_check_settings(check_settings).use_dom(True)

    @keyword("Send Dom")
    def send_dom(
        self,
        check_settings=None,  # type:Optional[SeleniumCheckSettings]
    ):
        # type: (...)->SeleniumCheckSettings
        """
        Returns a CheckSettings object specified send dom.

        *Example:*
            | Eyes Check   |    Target Window    |  Send Dom   |
        """
        return new_or_cur_check_settings(check_settings).send_dom(True)



from __future__ import absolute_import, unicode_literals

from typing import Optional, Text

from robot.libraries.BuiltIn import BuiltIn
from selenium.webdriver.remote.webdriver import WebDriver

from applitools.common import BatchInfo, MatchLevel, TestResults, TestResultsSummary
from applitools.selenium import Eyes

from ..base import LibraryComponent, keyword
from ..utils import get_enum_by_name, parse_viewport_size
from .keyword_tags import CHECK_FLOW


class RunnerKeywords(LibraryComponent):
    @keyword("Eyes Get All Test Results")
    def get_all_tests_results(self):
        # type: () -> TestResultsSummary
        try:
            test_results = self.eyes_runner.get_all_test_results()
            self.log_to_console(test_results)
            return test_results
        finally:
            # after fetching of results we don't need runner
            self.ctx.clean_eyes_runner()


class SessionKeywords(LibraryComponent):
    @keyword(  # noqa
        "Eyes Open",
        types={
            "app_name": (str, None),
            "test_name": (str, None),
            "viewport_size": (str, None),
            "host_os": (str, None),
            "host_app": (str, None),
            "match_level": (str, None),
            "batch": (str, BatchInfo, None),
            "baseline_env_name": (str, None),
            "hide_scrollbars": (bool, None),
            "stitch_mode": (str, None),
            "force_full_page_screenshot": (bool, None),
            "match_timeout": (int, None),
            "save_new_tests": (bool, None),
            "wait_before_screenshots": (bool, None),
            "send_dom": (bool, None),
            "is_disabled": (bool, None),
        },
        tags=(CHECK_FLOW,),
    )
    def open(  # noqa
        self,
        test_name=None,  # type: Optional[Text]
        viewport_size=None,  # type: Optional[Text]
        app_name=None,  # type: Optional[Text]
        host_os=None,  # type: Optional[Text]
        host_app=None,  # type: Optional[Text]
        match_level=None,  # type: Optional[Text]
        baseline_env_name=None,  # type: Optional[Text]
        batch=None,  # type: Optional[Text]
        branch_name=None,  # type: Optional[Text]
        parent_branch_name=None,  # type: Optional[Text]
        force_full_page_screenshot=None,  # type: Optional[bool]
        stitch_mode=None,  # type: Optional[Text]
        match_timeout=None,  # type: Optional[int]
        hide_scrollbars=None,  # type: Optional[bool]
        save_new_tests=None,  # type: Optional[bool]
        wait_before_screenshots=None,  # type: Optional[bool]
        send_dom=None,  # type: Optional[bool]
        is_disabled=None,  # type: Optional[bool]
    ):
        # type: (...)->WebDriver
        """
        Shared parameters section from `applitools.yaml` could be overwritten during `Eyes Open` call, see `Preconditions`.

            | =Arguments=                  | =Description=                                                                                                                               |
            | Test Name                    | By default fetched from name of current test. Could be overfritet here.                                                                     |
            | Viewport Size                | The viewport size of the browser window in format [width height] e.g. [1900 1080]                                                           |
            | App Name                     | The name of the application under test                                                                                                      |
            | Host OS                      | The operating system of the test, can be used to override the OS name to allow cross OS verification                                        |
            | Host App                     | The browser name for the test, can be used to override the browser name to allow cross browser verification                                 |
            | Match Level                  | The match level for the comparison of this test's checkpoints - can be STRICT, LAYOUT, CONTENT or EXACT                                     |
            | Baseline Env Name            | Name of the branch where the baseline reference will be taken from and where new and accepted steps will be saved to                        |
            | Batch                        | Accepts desired batch returned by `Create Batch Info` or batch name                                                                         |
            | Branch Name                  | The branch to use to check test                                                                                                             |
            | Parent Branch Name           | Parent Branch to base the new Branch on                                                                                                     |
            | Force Full Page Screenshot   | Will force the browser to take a screenshot of whole page                                                                                   |
            | Stitch Mode                  | Type of stitching used for full page screenshots - can be CSS or SCROLL                                                                     |
            | Match Timeout                | Determines how much time in milliseconds Eyes continues to retry the matching before declaring a mismatch on this test checkpoints          |
            | Hide Scrollbars              | Sets if the scrollbars are hidden this session's tests, by passing 'True' or 'False' in the variable                                        |
            | Save New Tests               | Sets if the new checkpoints on this session are automatically accepted, by passing 'True' or 'False' in the variable                        |
            | Wait Before Screenshots      | Determines the number of milliseconds that Eyes will wait before capturing a screenshot on this test checkpoints                            |
            | Send DOM                     | Sets if DOM information should be sent for this session's checkpoints                                                                       |
            | Is Disabled                  | Determines whether or not interactions with Eyes will be silently ignored for this test                                                     |

        *Example:*
            | Eyes Open | TestName | [1024 768] | AppNameOverride | OSOverride | HostAppOverride | batchname=Some batch name |
        """
        config_cloned = self.get_configuration()
        if app_name:
            config_cloned.app_name = app_name
        elif config_cloned.app_name is None:
            raise ValueError("app_name should be provided")

        if test_name:
            config_cloned.test_name = test_name
        elif config_cloned.test_name is None:
            # no test_name present use test name from the suite
            config_cloned.test_name = BuiltIn().get_variable_value("${TEST NAME}")

        if batch:
            config_cloned.batch = self.ctx.register_or_get_batch(batch)

        if viewport_size:
            config_cloned.viewport_size = parse_viewport_size(viewport_size)

        if host_os:
            config_cloned.host_os = host_os
        if host_app:
            config_cloned.host_app = host_app
        if match_level:
            config_cloned.match_level = get_enum_by_name(match_level, MatchLevel)
        if baseline_env_name:
            config_cloned.baseline_env_name = baseline_env_name
        if branch_name:
            config_cloned.branch_name = branch_name
        if parent_branch_name:
            config_cloned.parent_branch_name = parent_branch_name
        if force_full_page_screenshot:
            config_cloned.force_full_page_screenshot = force_full_page_screenshot
        if stitch_mode:
            config_cloned.stitch_mode = stitch_mode
        if match_timeout:
            config_cloned.match_timeout = match_timeout
        if hide_scrollbars:
            config_cloned.hide_scrollbars = hide_scrollbars
        if save_new_tests:
            config_cloned.save_new_tests = save_new_tests
        if wait_before_screenshots:
            config_cloned.wait_before_screenshots = wait_before_screenshots
        if send_dom:
            config_cloned.send_dom = send_dom
        if is_disabled:
            config_cloned.is_disabled = is_disabled

        eyes = Eyes(self.eyes_runner)
        eyes.set_configuration(config_cloned)
        self.register_eyes(eyes)
        return eyes.open(self.fetch_driver())

    @keyword(
        "Eyes Close Async",
        types=(bool,),
        tags=(CHECK_FLOW,),
    )
    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Closes a session and returns the results of the session.
        If a test is running, aborts it. Otherwise, does nothing.

            | =Arguments=                  | =Description=                                                                                                           |
            | Raise Exception (bool)       | If you don't want an exception to be thrown if there are new, missing or mismatched steps, pass 'False' in the variable |
        *Example:*
            | Eyes Close Async | ${false} |
        """
        # TODO: proper handle situation when we ask to raise an exception
        return self.current_eyes.close_async()

    @keyword(
        "Eyes Abort Async",
        tags=(CHECK_FLOW,),
    )
    def abort(self):
        # type: () -> Optional[TestResults]
        """
        Stops execution without calling close().
        This method does all the cleanup normally done by close.
        If this method is called, and close has not been called, then the test will
        have a status of Aborted in the Test Manager.

        *Example:*
            | Eyes Abort Async |
        """
        return self.current_eyes.abort_async()

    @keyword("Is Eyes Open")
    def is_open(self):
        # type: () -> bool
        """
        Returns True if Eyes is opened

        *Example:*
            | ${is_open}= | Is Eyes Open |
        """
        if self.current_eyes:
            return self.current_eyes.is_open
        return False



from __future__ import absolute_import, unicode_literals

from datetime import datetime
from typing import Any, Optional, Text, Union

from robot.api.deco import keyword
from six import string_types as basestring

from applitools.selenium import BatchInfo

from ..base import LibraryComponent


class ConfigurationKeywords(LibraryComponent):
    @keyword(
        "Create Batch Info",
        types={
            "name": str,
            "started_at": (datetime, str, None),
            "batch_sequence_name": (str, None),
            "batch_id": (str, None),
        },
    )
    def create_batch_info(
        self,
        name,  # type: Text
        started_at=None,  # type: Optional[Union[datetime,Text]]
        batch_sequence_name=None,  # type: Optional[Text]
        batch_id=None,  # type: Optional[Text]
    ):
        # type: (...) -> Text
        """
        Returns a BatchInfo ID string that may be used as batch argument on `Eyes Open`.

            | =Arguments=                  | =Description=                                                                              |
            | Name                         | The name of the batch                                                                      |
            | Started At                   | The date and time that will be displayed in the Test Manager as the batch start time *(*)* |
            | Batch ID                     | This argument groups together tests ran in different executions                            |

        The *Started At* argument may be passed as:
        - String: YYYY-mm-dd HH:MM:SS
        - Datetime variable: See [https://robotframework.org/robotframework/latest/libraries/DateTime.html|DateTime library]

        *Example:*
            | ${batch_id}= | Create Eyes Batch |
        """

        if started_at:
            if isinstance(started_at, basestring):
                started_at = datetime.strptime(started_at, "%Y-%m-%d %H:%M:%S")
            elif not isinstance(started_at, datetime):
                raise TypeError("BatchInfo started_at should be `str` or `datetime`")
        batch = BatchInfo(
            name, started_at=started_at, batch_sequence_name=batch_sequence_name
        )
        if batch_id:
            batch = batch.with_batch_id(batch_id)
        self.ctx.register_or_get_batch(batch)
        return batch.id

    @keyword("Get Eyes Configure Property", types=(str,))
    def get_eyes_configure_property(self, name):
        # type: (Text) -> Any
        return getattr(self.ctx.configure, name)

    @keyword("Set Eyes Configure Property", types=(str, str))
    def set_eyes_configure_property(self, name, value):
        # type: (Text, Any) -> Any
        setattr(self.ctx.configure, name, value)



from __future__ import absolute_import, unicode_literals

from typing import TYPE_CHECKING, Any, Text

from appium.webdriver import WebElement as AppiumWebElement
from robot.api.deco import keyword as original_keyword
from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement

from applitools.selenium import Target
from applitools.selenium.fluent import SeleniumCheckSettings

from ..base import LibraryComponent
from ..utils import collect_check_settings, parse_region
from .keyword_tags import CHECK_SETTINGS_SUPPORT, TARGET_KEYWORD

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import AnyWebElement

    from ..custom_types import Locator


def keyword(name=None, tags=(), types=()):
    """Keyword with predefined CHECK_SETTINGS_SUPPORT tag"""
    tags = tags + (CHECK_SETTINGS_SUPPORT, TARGET_KEYWORD)
    return original_keyword(name, tags, types)


class TargetKeywords(LibraryComponent):
    @keyword("Target Window")
    def target_window(self, *check_settings_keywords):
        # type: (tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with Window selected and any number of Check Settings Keywords.

        *Example:*
            |  ${target}=    | Target Window     |
        """
        return collect_check_settings(
            Target.window(), self.defined_keywords, *check_settings_keywords
        )

    @keyword(
        "Target Region By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
    )
    def target_region_by_element(self, element, *check_settings_keywords):
        # type: (AnyWebElement,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Region and any number of Check Settings Keywords.

            |  =Arguments=  | =Description=                                          |
            | Element       | *Mandatory* - The element to check                     |

        *Example:*
            |  ${target}=  |  Target Region By Element  |  ${element}  |
        """
        return collect_check_settings(
            Target.region(element), self.defined_keywords, *check_settings_keywords
        )

    @keyword(
        "Target Region By Coordinates", types=(str,), tags=(CHECK_SETTINGS_SUPPORT,)
    )
    def target_region_by_coordinates(self, region, *check_settings_keywords):
        # type: (Text,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Region and any number of Check Settings Keywords.

            |  =Arguments=  | =Description=                                                       |
            |  Region       | *Mandatory* - The region to check in format [left top width height] ,e.g. [100 200 300 300]  |

        *Example:*
            |  ${target}=  |  Target Region By Coordinates  |  [10 30 40 50]  |
        """
        return collect_check_settings(
            Target.region(parse_region(region)),
            self.defined_keywords,
            *check_settings_keywords
        )

    @keyword("Target Region By Selector", types=(str,), tags=(CHECK_SETTINGS_SUPPORT,))
    def target_region_by_selector(self, selector, *check_settings_keywords):
        # type: (Text,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Region and any number of Check Settings Keywords.

            | =Arguments=   | =Description=                          |
            |  Selector     | *Mandatory* - The selector to check.   |

        *Example:*
            |  ${target}=  |  Target Frame By Selector  |  css:#selector  |
        """
        return collect_check_settings(
            Target.region(self.from_locator_to_supported_form(selector)),
            self.defined_keywords,
            *check_settings_keywords
        )

    @keyword(
        "Target Frame By Element",
        types={"element": (SeleniumWebElement, AppiumWebElement)},
        tags=(CHECK_SETTINGS_SUPPORT,),
    )
    def target_frame_by_element(self, element, *check_settings_keywords):
        # type: (AnyWebElement,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Frame and any number of Check Settings Keywords.

            | =Arguments=   | =Description=                                        |
            | Element       | *Mandatory* - The frame to check                     |

        *Example:*
            |  ${target}=  |  Target Frame By Element  |  ${element}  |
        """
        return collect_check_settings(
            Target.frame(element), self.defined_keywords, *check_settings_keywords
        )

    @keyword(
        "Target Frame By Selector",
        types=(str,),
        tags=(CHECK_SETTINGS_SUPPORT,),
    )
    def target_frame_by_selector(self, selector, *check_settings_keywords):
        # type: (Locator,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Frame and any number of Check Settings Keywords.

            | =Arguments=   | =Description=                                  |
            |  Selector     | *Mandatory* - Selector of the frame to check.  |

        *Example:*
            |  ${target}=  |  Target Frame By Selector  |  css:#selector  |
        """
        return collect_check_settings(
            Target.frame(self.from_locator_to_supported_form(selector)),
            self.defined_keywords,
            *check_settings_keywords
        )

    @keyword(
        "Target Frame By Index",
        types=(int,),
        tags=(CHECK_SETTINGS_SUPPORT,),
    )
    def target_frame_by_index(self, frame_index, *check_settings_keywords):
        # type: (int,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Frame and any number of Check Settings Keywords.

            | =Arguments=     | =Description=                                   |
            |  Frame Index    | *Mandatory* - Index of the frame to check.      |

        *Example:*
            | ${target}=  |  Target Frame By Index  |  2  |
        """
        return collect_check_settings(
            Target.frame(frame_index), self.defined_keywords, *check_settings_keywords
        )

    @keyword(
        "Target Frame By Name",
        types=(str,),
        tags=(CHECK_SETTINGS_SUPPORT,),
    )
    def target_frame_by_name(self, frame_name, *check_settings_keywords):
        # type: (Text,tuple[Any]) -> SeleniumCheckSettings
        """
        Returns a CheckSettings object with selected Frame and any number of Check Settings Keywords.

            |  =Arguments=   | =Description=                                  |
            |  Frame Name    | *Mandatory* - Name of the frame to check.      |

        *Example:*
            |  ${target}=  |  Target Frame By Name  |  frameName  |
        """
        return collect_check_settings(
            Target.frame(frame_name), self.defined_keywords, *check_settings_keywords
        )



from argparse import ArgumentParser


def init_config_command(_):
    from . import utils

    utils.copy_config_to(".")


def cli_parser():
    parser = ArgumentParser(
        prog="python -m EyesLibrary",
        description="Applitools Eyes Robot Framework support library",
    )
    parser.set_defaults(command=lambda _: parser.print_help())
    commands = parser.add_subparsers()
    init_config = commands.add_parser(
        "init-config",
        description="Create default config file",
        help="Create default config file",
    )
    init_config.set_defaults(command=init_config_command)
    return parser


args = cli_parser().parse_args()
args.command(args)



from typing import Optional

import attr

from applitools.common.selenium import Configuration
from applitools.selenium import RunnerOptions


@attr.s
class RobotConfiguration(Configuration):
    runner_options = attr.ib(default=None)  # type: Optional[RunnerOptions]



from __future__ import absolute_import, unicode_literals

import os
import re
import shutil
from collections import OrderedDict
from enum import Enum
from typing import Any, Generator, Text, Type

from robot.libraries.BuiltIn import BuiltIn

from applitools.common import RectangleSize, Region
from applitools.common.utils import argument_guard
from applitools.common.validators import is_webelement
from applitools.selenium.fluent import SeleniumCheckSettings

SEPARATOR = object()


def extract_keyword_and_arguments(
    keywords_from_test,  # type: list[Any]|tuple[Any]
    defined_keywords,  # type: list[Any]|tuple[Any]
):
    # type: (...) -> Generator[tuple[Text, list[Any]], None, None]
    res = OrderedDict()
    key_keyword = None
    key_keyword_index = -1
    for i, keyword in enumerate(keywords_from_test):
        if keyword == key_keyword and key_keyword_index != i:
            # Next keyword with similar name
            res[key_keyword].append(SEPARATOR)

        if keyword in defined_keywords:
            key_keyword = keyword
            key_keyword_index = i
            if key_keyword not in res:
                res[key_keyword] = []
            continue

        if key_keyword is None:
            raise ValueError(
                "Incorrect keyword argument. Keywords: {}".format(keywords_from_test)
            )
        res[key_keyword].append(keyword)
    for keyword, arguments in res.items():
        yield keyword, arguments


def splits_args_by_separator(args):
    # type: (list[Any]) -> Generator[tuple[Any], None, None]
    res = ()
    for i, arg in enumerate(args, start=1):
        if arg is not SEPARATOR:
            res += (arg,)
        else:
            yield res
            res = ()
            continue

        if len(args) <= i:
            yield res


def collect_check_settings(check_settings, defined_keywords, *keywords):
    # type: (SeleniumCheckSettings,list[str],tuple[Any])->SeleniumCheckSettings
    """ Fill `check_setting` with data from keyword and return `check_settings`"""
    for keyword, keyword_args in extract_keyword_and_arguments(
        keywords, defined_keywords
    ):
        if keyword_args:
            # keyword has arguments
            for separated_args in splits_args_by_separator(keyword_args):
                separated_args += (check_settings,)
                BuiltIn().run_keyword(keyword, *separated_args)
        else:
            # in case keyword without args like `Fully`
            BuiltIn().run_keyword(keyword, check_settings)
    return check_settings


int_float_pattern = r"\d+(?:\.\d+)?"


def parse_viewport_size(text):
    # type: (Text) -> RectangleSize
    num_ptrs = (int_float_pattern,) * 2
    match = re.match(r"\[(%s) (%s)\]" % num_ptrs, text)
    if match is None:
        raise ValueError(
            "Incorrect value of viewport: {}.\n\t Format should be: [800 700]".format(
                text
            )
        )
    groups = match.groups()
    return RectangleSize(width=float(groups[0]), height=float(groups[1]))


def parse_region(text):
    num_ptrs = (int_float_pattern,) * 4
    match = re.match(r"\[(%s) (%s) (%s) (%s)\]" % num_ptrs, text)
    if match is None:
        raise ValueError(
            "Incorrect value of region: {}.\n\t Format should be: [10 10 10 10]".format(
                text
            )
        )
    groups = match.groups()
    return Region(
        left=float(groups[0]),
        top=float(groups[1]),
        width=float(groups[2]),
        height=float(groups[3]),
    )


def is_webelement_guard(element):
    argument_guard.is_valid_type(
        is_webelement(element),
        "element argument should be type Selenium or Appium Web Element",
    )


def get_config_file_path():
    here = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(here, "applitools.yaml")


def copy_config_to(path_to_dir):
    if not os.path.exists(path_to_dir):
        raise ValueError("Directory doesn't exists")
    return shutil.copy(get_config_file_path(), path_to_dir)


def get_enum_by_name(name, enm):
    # type: (Text, Type[Enum]) -> Enum
    try:
        return getattr(enm, name.upper())
    except AttributeError:
        raise ValueError("`{}` does not contain `{}`".format(enm, name))



# #!/usr/bin/env python
import codecs
import re
import sys
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))


def read(filename):
    return codecs.open(path.join(here, filename), "r", "utf-8").read()


# preventing ModuleNotFoundError caused by importing lib before installing deps
def get_version():
    version_file = read("src/EyesLibrary/__version__.py")
    try:
        version = re.findall(r"^__version__ = \"([^']+)\"\r?$", version_file, re.M)[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

    return version


install_requires = [
    "trafaret == 2.1.0",
    "PyYAML >= 5,<6",
    "eyes_selenium >= 4.21.0",
]
if sys.version_info[:2] <= (2, 7):
    install_requires.extend(
        [
            "robotframework >= 3.2.2,<4.0 ",
            "robotframework-pythonlibcore >= 2.0.2,<3.0",
            "robotframework-seleniumlibrary >= 3.3.1,<4.5",
            "robotframework-appiumlibrary >= 1.5,<1.6",
        ]
    )
else:
    install_requires.extend(
        [
            "robotframework >= 4.0 ",
            "robotframework-pythonlibcore >= 3.0",
            "robotframework-seleniumlibrary >= 4.5",
            "robotframework-appiumlibrary >= 1.6",
        ]
    )

setup(
    name="eyes-robotframework",
    version=get_version(),
    url="http://www.applitools.com",
    author="Applitools Team",
    author_email="team@applitools.com",
    description="Applitools Python SDK. Robot Framework package",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    python_requires=">=2.7, <4",
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Framework :: Robot Framework :: Library",
    ],
    keywords="applitools robotframework eyes eyes_selenium selenium appium",
    install_requires=install_requires,
    package_dir={"": "src"},
    packages=find_packages("src"),
    package_data={
        "": ["README.rst", "LICENSE"],
        "EyesLibrary": ["py.typed", "applitools.yaml"],
    },
    project_urls={
        "Bug Reports": "https://github.com/applitools/eyes.sdk.python/issues",
        "Source": "https://github.com/applitools/eyes.sdk.python/tree/master"
        "/eyes_robotframework",
    },
)



CHECK_SETTING = "Check Settings keyword"
UFG_RELATED = "Ultrafast Grid only"
TARGET_SUPPORT = "Allow to pass Target keyword"
CHECK_SETTINGS_SUPPORT = "Allow to pass Check Settings keywords"
CHECK_FLOW = "Check flow"
TARGET_KEYWORD = "Target Keyword"



from __future__ import absolute_import, unicode_literals

from robot.api import logger
from robot.utils import ConnectionCache, is_string

from applitools.selenium import Eyes


def is_noney(item):
    return item is None or is_string(item) and item.upper() == "NONE"


class EyesCache(ConnectionCache):
    def __init__(self):
        ConnectionCache.__init__(self, no_current_msg="No current eyes")
        self._closed = set()

    @property
    def eyes(self):
        # type: () -> Eyes
        return self._connections

    @property
    def active_eyes(self):
        open_eyes = []
        for eyes in self._connections:
            if eyes not in self._closed:
                open_eyes.append(eyes)
        return open_eyes

    @property
    def active_eyes_ids(self):
        open_eyes_ids = []
        for index, driver in enumerate(self._connections):
            if driver not in self._closed:
                open_eyes_ids.append(index + 1)
        return open_eyes_ids

    @property
    def active_aliases(self):
        return self._aliases

    def close(self):
        if not self.current:
            return
        eyes = self.current
        error = self._quit(eyes, None)
        for alias in self._aliases:
            if self._aliases[alias] == self.current_index:
                del self._aliases[alias]
        self.current = self._no_current
        self._closed.add(eyes)
        if error:
            raise error

    def close_all(self):
        error = None
        for eyes in self._connections:
            if eyes not in self._closed:
                error = self._quit(eyes, error)
        self.empty_cache()
        if error:
            raise error
        return self.current

    def _quit(self, driver, error):
        try:
            driver.quit()
        except Exception as exception:
            logger.error("When closing browser, received exception: %s" % exception)
            error = exception
        return error

    def get_index(self, alias_or_index):
        index = self._get_index(alias_or_index)
        try:
            driver = self.get_connection(index)
        except RuntimeError:
            return None
        return None if driver in self._closed else index

    def _get_index(self, alias_or_index):
        alias_or_index = None if is_noney(alias_or_index) else alias_or_index
        try:
            return self.resolve_alias_or_index(alias_or_index)
        except AttributeError:
            pass
        except ValueError:
            return None
        # TODO: This try/except block can be removed when minimum
        #  required Robot Framework version is 3.3 or greater.
        try:
            return self._resolve_alias_or_index(alias_or_index)
        except ValueError:
            return None
